const axios = require('axios')
const fs = require('fs');
const { fake_mail_address, get_message } = require('./mail.js');
const { log, sleep, play_sound, log_with_req_id, now } = require('./utils.js');
const cheerio = require('cheerio');
const prompt = require('prompt-sync')();
const { v4: uuidv4 } = require('uuid');

// const proxy = {
//     protocol: 'http',
//     host: '127.0.0.1',
//     port: 7890,
// };

const proxy = undefined;
const userAgent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.61 Safari/537.36';

function load_categories() {
    const data = JSON.parse(fs.readFileSync("data/categories.json", { encoding: 'utf8', flag: 'r' }));
    return data;
}

async function get_index(code, categories, page, cookie) {
    // 'categories': '第1册',
    // 'categories': '第1册,一部',
    // 'categories': '第1册,一部,一',
    // 'dictionaryCode': 'hydcdcx',
    log(`开始抓取索引: code: ${code}, categores: ${categories}, page: ${page}`);

    const params = {
        'dictionaryCode': code,
        'categories': categories,
        'page': page
    };

    const headers = {
        'Accept': '*/*',
        'Accept-Language': 'zh-CN,zh;q=0.9',
        'Ajax-Referrer': '/dictionary/index?dictionaryCode=hydcdcx&categories=&page=0',
        'Ajax-Request': 'true',
        'Cache-Control': 'no-cache',
        'Connection': 'keep-alive',
        'DNT': '1',
        'Host': 'www.hanyucidian.org',
        'Pragma': 'no-cache',
        'Referer': 'https://www.hanyucidian.org/dictionary/index?dictionaryCode=hydcdcx&categories=&page=0',
        'Sec-Fetch-Dest': 'empty',
        'Sec-Fetch-Mode': 'cors',
        'Sec-Fetch-Site': 'same-origin',
        'User-Agent': userAgent,
        'Cookie': cookie
    };

    return axios.get('https://www.hanyucidian.org/dictionary/index', { params, headers, proxy }).then(res => {
        log(`索引抓取成功`);
        return res.data;
    }).catch(error => {
        log(`索引抓取失败: ${error.code}`);
        return Promise.reject(error);
    });
}

async function get_all_indexes_and_entries(code, entry, delay_ms, only_index, cookie) {
    const path = 'data/' + code + '/' + entry + '.json';
    let new_items = {};

    log(`检查索引文件路径: ${path}`);
    if (fs.existsSync(path)) {
        log(`索引已存在，直接加载索引`);
        const data = JSON.parse(fs.readFileSync(path, { encoding: 'utf8', flag: 'r' }));
        new_items = data.data.items;
    } else {
        log(`索引不存在，尝试抓取索引...`);

        await sleep(delay_ms);
        const data = await get_index(code, entry, 0, cookie);
        const pages = parseInt(data.pages);

        log(`索引分页: ${pages}`);

        if (pages == 1) {
            log(`索引写入文件`)
            fs.writeFileSync(path, JSON.stringify({ entry, data }));
            new_items = data.items;
        } else {
            let items = [];
            for (const page of [...Array(data.pages).keys()]) {
                if (page == 0) {
                    items = items.concat(data.items);
                } else {
                    await sleep(delay_ms);
                    const data2 = await get_index(code, entry, page, cookie);
                    items = items.concat(data2.items);
                }
            }

            log(`索引写入文件`)
            fs.writeFileSync(path, JSON.stringify({ entry, data: { pages, items } }));
            new_items = items;
        }
    }

    log(`循环处理子项任务: ${entry}`);

    for (const item of new_items) {
        log(`子项任务类型: ${item.type}`);
        if (item.type == "index") {
            await get_all_indexes_and_entries(code, entry + "," + item.entry, delay_ms, only_index, cookie);
        } else if (item.type == "entry") {
            if (only_index) {
                continue;
            } else {
                let path = `data/${code}/entries/${item.entry}.json`;
                log(`检查词条文件路径: ${path}`);

                if (fs.existsSync(path)) {
                    log(`词条已存在，跳过`);
                    continue;
                }

                log(`词条不存在，尝试抓取词条...`);

                await sleep(delay_ms);
                let data = await get_entry(code, item.entry, cookie);
                log(`词条写入文件`)
                fs.writeFileSync(path, JSON.stringify(data));
            }
        }
    }
}

async function get_entry(code, entry, cookie) {
    // 'dictionaryCode': 'hydcdcx',
    // 'entryName': '中'
    log(`开始抓取词条: code: ${code}, entry: ${entry}`);

    const params = {
        'dictionaryCode': code,
        'entryName': entry
    };

    const headers = {
        'Accept': '*/*',
        'Accept-Language': 'zh-CN,zh;q=0.9',
        'Ajax-Referrer': '/dictionary/index?dictionaryCode=hydcdcx&categories=%E7%AC%AC1%E5%86%8C%2C%E4%B8%A8%EF%BC%88%E4%BA%85%EF%BC%89%E9%83%A8%2C%E4%B8%AD&page=0',
        'Ajax-Request': 'true',
        'Cache-Control': 'no-cache',
        'Connection': 'keep-alive',
        'DNT': '1',
        'Pragma': 'no-cache',
        'Referer': 'https://www.hanyucidian.org/dictionary/index?dictionaryCode=hydcdcx&categories=%E7%AC%AC1%E5%86%8C%2C%E4%B8%A8%EF%BC%88%E4%BA%85%EF%BC%89%E9%83%A8%2C%E4%B8%AD&page=0',
        'Sec-Fetch-Dest': 'empty',
        'Sec-Fetch-Mode': 'cors',
        'Sec-Fetch-Site': 'same-origin',
        'User-Agent': userAgent,
        'Cookie': cookie
    }

    return axios.get('https://www.hanyucidian.org/dictionary/entry', { params, headers, proxy }).then(res => {
        log(`词条抓取成功`);
        return res.data;
    }).catch(error => {
        log(`词条抓取失败: ${error.code}`);
        return Promise.reject(error);
    });
}

async function get_dictionary(code, delay_ms, only_index, cookie) {
    const categories = load_categories()[code];
    const dir = `data/${code}/entries`;

    if (!fs.existsSync(dir)) {
        fs.mkdirSync(dir, { recursive: true });
    }

    for (const item of categories.data.items) {
        await get_all_indexes_and_entries(code, item.entry, delay_ms, only_index, cookie);
    }

    console.log(`[status]: done!`);
}

async function get_session() {
    const req_id = uuidv4();
    const log = log_with_req_id(req_id);

    const mail_address = fake_mail_address();
    log(`注册新账号: ${mail_address}`);

    let headers = {
        'Host': 'www.hanyucidian.org',
        'User-Agent': userAgent,
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
        'Accept-Language': 'zh-SG,en-US;q=0.7,en;q=0.3',
        'Referer': 'https://www.hanyucidian.org/registration/register',
        'Connection': 'keep-alive',
        'Upgrade-Insecure-Requests': '1',
        'Sec-Fetch-Dest': 'document',
        'Sec-Fetch-Mode': 'navigate',
        'Sec-Fetch-Site': 'same-origin',
        'Sec-Fetch-User': '?1',
    };

    let csrf = "";
    let resp_html = "";
    let has_captcha;
    let captcha_value;
    let cookie;

    try {
        log(`开始获取注册页面...`);
        let resp = await axios.get('https://www.hanyucidian.org/registration/register', { headers, proxy });
        resp_html = resp.data;
        log(`注册页面下载完成`);

        let $ = cheerio.load(resp_html);
        has_captcha = $('#captcha-element').length != 0;
        csrf = $('input[name=_csrf]').attr('value');
        log(`获取csrf: ${csrf}`);

        let id = resp.headers['set-cookie']
            .find(value => value.includes('JSESSIONID'))
            ?.match(new RegExp(`^JSESSIONID=(.+?);`))
            ?.[1];

        cookie = `JSESSIONID=${id}`;
        log(`获取到的cookie: ${cookie}`);
    } catch (error) {
        log(`下载注册页面失败: ${error.code}`);
        throw error;
    }

    if (has_captcha) {
        log(`开始获取计算器验证图片`);
        let buffer;

        try {
            const resp = await axios.get('https://www.hanyucidian.org/captcha', {
                headers: {
                    'Host': 'www.hanyucidian.org',
                    'User-Agent': userAgent,
                    'Accept': 'image/avif,image/webp,*/*',
                    'Accept-Language': 'zh-SG,en-US;q=0.7,en;q=0.3',
                    'Referer': 'https://www.hanyucidian.org/registration/register',
                    'Connection': 'keep-alive',
                    'Sec-Fetch-Dest': 'image',
                    'Sec-Fetch-Mode': 'no-cors',
                    'Sec-Fetch-Site': 'same-origin',
                    'Cookie': cookie,
                },
                responseType: 'arraybuffer',
                proxy
            });
            buffer = resp.data;
        } catch (error) {
            log(`下载计算器验证图片失败: ${error.message}`);
            throw error;
        }
        play_sound();

        const { default: display_image } = await import('terminal-image');
        console.log(await display_image.buffer(buffer, { width: '30%', height: '30%' }));
        fs.writeFileSync('captcha.png', buffer);

        let input = prompt('请输入计算的公式：').trim();
        captcha_value = input.split('=')[1];
        log(`接受输入: ${input}`);
        const filename = `captcha/${req_id}#${input}.png`;
        fs.writeFileSync(filename, buffer);
    }

    let password = Buffer.from(mail_address).toString('base64');
    log(`尝试提交注册信息: ${mail_address}: ${password}`);

    try {
        let forms = {
            '_csrf': csrf,
            'plan': '',
            'email': mail_address,
            'repeatedEmail': mail_address,
            'password': password,
            'repeatedPassword': password,
            'acceptAgb': 'true',
            '_acceptAgb': 'on',
            'captcha': captcha_value
        };

        let resp = await axios.post(
            'https://www.hanyucidian.org/registration/register',
            new URLSearchParams(forms),
            {
                headers: {
                    'Host': 'www.hanyucidian.org',
                    'User-Agent': userAgent,
                    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
                    'Accept-Language': 'zh-SG,en-US;q=0.7,en;q=0.3',
                    'Referer': 'https://www.hanyucidian.org/registration/register',
                    'Origin': 'https://www.hanyucidian.org',
                    'Connection': 'keep-alive',
                    'Upgrade-Insecure-Requests': '1',
                    'Sec-Fetch-Dest': 'document',
                    'Sec-Fetch-Mode': 'navigate',
                    'Sec-Fetch-Site': 'same-origin',
                    'Sec-Fetch-User': '?1',
                    'Cookie': cookie
                },
                proxy
            }
        );

        log(`提交注册响应状态: ${resp.status}`);
        if (resp.data.includes("请检查您的收件箱，并点击邮件中的链接继续")) {
            log(`注册成功: ${resp.status}`);
        }
    } catch (error) {
        log(`注册账号失败: ${error.message}`);
        throw error;
    }
    let register_url

    try {
        register_url = await get_message(req_id, mail_address);
        log(`验证链接: ${register_url}`);
    } catch (error) {
        log(`获取邮件验证链接失败: ${error.message}`);
        throw new Error('没有发现新邮件');
    }

    cookie = undefined;

    try {
        log(`尝试完成链接验证...`);

        let resp = await axios.get(register_url, {
            headers: {
                'User-Agent': userAgent,
                'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
                'Accept-Language': 'zh-SG,en-US;q=0.7,en;q=0.3',
                'Accept-Encoding': 'gzip, deflate, br',
                'Referer': 'https://www.hanyucidian.org/login',
                'Connection': 'keep-alive',
                'Upgrade-Insecure-Requests': '1',
                'Sec-Fetch-Dest': 'document',
                'Sec-Fetch-Mode': 'navigate',
                'Sec-Fetch-Site': 'same-origin'
            },
            proxy
        });

        let id = resp.headers['set-cookie']
            .find(value => value.includes('JSESSIONID'))
            ?.match(new RegExp(`^JSESSIONID=(.+?);`))
            ?.[1];
        cookie = `JSESSIONID=${id}`;
        log(`完成邮箱链接验证: ${mail_address}: ${cookie}`);
    } catch (error) {
        log(`邮箱链接验证失败: ${error.message}`);
        throw new Error('邮箱链接验证失败');
    }

    return { 'id': req_id, mail_address, password, cookie, date: now() };
}

async function run_spider(code, delay_ms, only_index, cookie) {
    const req_id = uuidv4();
    const log = log_with_req_id(req_id);

    if (!cookie) {
        while (true) {
            try {
                let session = await get_session();
                cookie = session;
                break;
            } catch (error) {
                log(`会话身份获取失败: ${error.message}`)
            }
        }
    }

    while (true) {
        try {
            await get_dictionary(code, delay_ms, only_index, cookie);
        } catch (error) {
            log(`读取词典失败: ${error.message}`)
            return;
        }
    }
}

// code: 'hydcdcx' 词典代码
// delay_ms: 2000 访问间隔，单位毫秒
// only_index: true 只抓取索引, false 抓取索引和词条
// cookie: JSESSIONID=0E84D6684B1A5656BD2E8EF9E67408B1 指定 cookie，跳过自动注册
run_spider('hydcdcx', 2000, true).catch(err => { console.log(err) });