const { faker } = require('@faker-js/faker');
const { log_with_req_id, sleep } = require('./utils.js');
const axios = require('axios');
const fs = require('fs');

const domains = ["1secmail.com", "1secmail.org", "1secmail.net", "icznn.com", "ezztt.com", "vjuum.com", "laafd.com", "txcct.com"];

function random_int(min, max) {
    return Math.floor(Math.random() * (max - min)) + min;
}

function fake_mail_address() {
    return faker.internet.exampleEmail().toLowerCase().split('@')[0] + "@" + domains[random_int(0, domains.length)]
}

async function get_message(req_id, mail_address) {
    const log = log_with_req_id(req_id);
    const [name, domain] = mail_address.split('@');
    let max_limit = 0;
    let id = 0;

    while (true) {
        max_limit += 1;

        if (max_limit > 2) {
            throw new Error(`没有发现新邮件: 已尝试 ${max_limit} 次`);
        }

        log(`获取邮件内容: ${mail_address}`);

        try {
            const new_url = `https://www.1secmail.com/api/v1/?action=getMessages&login=${name}&domain=${domain}`;
            const resp = await axios.get(new_url);
            if (resp.data.length > 0) {
                if (resp.data[0]["subject"].includes("电子邮件确认和帐户注册")) {
                    id = resp.data[0]["id"];
                    break;
                }
            } else {
                log(`没有新邮件`);
            }
        } catch (error) {
            log(`获取邮件列表失败: ${mail_address}: ${error.message}`);
            throw error;
        }

        await sleep(5 * 1000); // 等待10秒
    }

    try {
        const msg_url = `https://www.1secmail.com/api/v1/?action=readMessage&login=${name}&domain=${domain}&id=${id}`;
        const resp = await axios.get(msg_url);
        if (resp.data) {
            let msg = resp.data["textBody"];
            for (let line of msg.split('\n')) {
                let url = line.trim();
                if (url.startsWith("https://www.hanyucidian.org/registration/finish/")) {
                    return url;
                }
            }
        }
    } catch (error) {
        log(`获取指定消息失败: ${mail_address}: ${id}: ${error.message}`);
        throw error;
    }
}

exports.fake_mail_address = fake_mail_address;
exports.get_message = get_message;